import bpy
import rna_keymap_ui
import textwrap

from mathutils import Matrix
from bpy.props import EnumProperty

from .cursor_plus import updatePanel
from .cursor_undo import toggleUndo
from .cursor_gizmo import toggleGizmo
from . import cursor_text
from . import preset_management


pie_items = [
    ("MOVECURSOR", "Snap 3D Cursor", ""),
    ("MOVE", "Move", ""),
    ("ROTATE", "Rotate", ""),
    ("SCALE", "Scale", ""),
    ("TWEAK", "Tweak Snapping", ""),
    ("CLEAR", "Clear Rotation", ""),
    ("COPY", "Copy Rotation", ""),
    ("OVER", "Override", ""),
    ("PRESET", "Gizmo Preset", ""),
    ("EMPTY", "- Not used - ", ""),
]

plus_ops = (
    "object.cursor_plus_snap",
    "object.cursor_plus_clear",
    "object.cursor_plus_move",
    "object.cursor_plus_rotate",
    "object.cursor_plus_scale",
    "object.cursor_plus_copy"
)

txt = cursor_text

# Multiline text output. Found this workaround on b3d.interplanety.org. Huge thanks to Nikita!
def _label_multiline(context, text, parent):
    scale = context.preferences.system.ui_scale 
    chars = int(context.region.width/(6.2*scale))   
    wrapper = textwrap.TextWrapper(width=chars)
    text_lines = wrapper.wrap(text=text)
    for text_line in text_lines:
        parent.label(text=text_line)

def redraw_gizmo(self, context):
    for window in bpy.context.window_manager.windows:
        for area in window.screen.areas:
            if area.type == 'VIEW_3D':
                with bpy.context.temp_override(area=area):
                    bpy.context.area.tag_redraw()


class CURSORPLUS_OT_gizmo_presets(bpy.types.Operator):
    bl_idname = "cursorplus.gizmo_presets"
    bl_label = "3D Cursor Gizmo"
    bl_options = {'REGISTER', 'INTERNAL'}
    bl_description = "Toggle between Gizmo Presets"
    
    def execute(self, context):
        prefs = context.preferences.addons[__package__].preferences
        prefs.preset_select = not prefs.preset_select
        return {'FINISHED'}


class CursorPlusPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    pref_tabs: EnumProperty(
        name = "Category",
        description="Preferences Tabs",
        items = [
            ('GIZMO', "Axes / Gizmo", ""),
            ('PLUS_PREFS', "Keymap", ""),
            ('UNDO', "Undo Info", ""),
        ],
        default = 'GIZMO'
    )
    
# Pie items   Pie items   Pie items   Pie items   Pie items   Pie items   Pie items   Pie items  

    pie_left: EnumProperty(
        name = "Left",
        items = pie_items,
        default = "MOVECURSOR"
    )
    
    pie_right: EnumProperty(
        name = "Right",
        items = pie_items,
        default = "MOVE"
    )

    pie_bottom: EnumProperty(
        name = "Bottom",
        items = pie_items,
        default = "ROTATE"
    )
    
    pie_top: EnumProperty(
        name = "Top",
        items = pie_items,
        default = "SCALE"
    )
    
    pie_top_left: EnumProperty(
        name = "Top Left",
        items = pie_items,
        default = "TWEAK"
    )
    
    pie_top_right: EnumProperty(
        name = "Top Right",
        items = pie_items,
        default = "OVER"
    )
   
    pie_bottom_left: EnumProperty(
        name = "Bottom Left",
        items = pie_items,
        default = "PRESET"
    )
  
    pie_bottom_right: EnumProperty(
        name = "Bottom Right",
        items = pie_items,
        default = "COPY"
    )

# Gizmo prefs   Gizmo prefs   Gizmo prefs   Gizmo prefs   Gizmo prefs   Gizmo prefs   Gizmo prefs    

    use_3d_cursor_gizmo: bpy.props.BoolProperty(
        name = "Use 3D Cursor Gizmo",
        default = True,
        update = toggleGizmo
    )

    show_axes: bpy.props.BoolProperty(
        name = "Show Axes",
        default = True
    )
    
    show_axes2: bpy.props.BoolProperty(
        name = "Show Axes",
        default = True
    )

    axes_gizmo: bpy.props.BoolProperty(
        name = "Axes as Gizmo",
        default = True
    )
    
    axes_gizmo2: bpy.props.BoolProperty(
        name = "Axes as Gizmo",
        default = True
    )

    show_dot: bpy.props.BoolProperty(
        name = "Show Dot",
        default = True
    )
    
    show_dot2: bpy.props.BoolProperty(
        name = "Show Dot",
        default = True
    )
    

    dot_gizmo: bpy.props.BoolProperty(
        name = "Dot as Gizmo",
        default = True
    )
    
    dot_gizmo2: bpy.props.BoolProperty(
        name = "Dot as Gizmo",
        default = True
    )

    custom_gizmo_size: bpy.props.FloatProperty(
        name  ="Custom Gizmo Size",
        default = 0.6,
        min = 0.0,
        soft_max = 3.0,
        subtype = 'FACTOR'
    )
    
    custom_gizmo_size2: bpy.props.FloatProperty(
        name  ="Custom Gizmo Size",
        default = 0.6,
        min = 0.0,
        soft_max = 3.0,
        subtype = 'FACTOR'
    )

    arrow_length: bpy.props.FloatProperty(
        name = "Arrow Length",
        default = 0.8,
        min = 0.0,
        soft_max = 10.0,
        subtype = 'FACTOR'
    )
    
    arrow_length2: bpy.props.FloatProperty(
        name = "Arrow Length",
        default = 0.8,
        min = 0.0,
        soft_max = 10.0,
        subtype = 'FACTOR'
    )

    arrow_width: bpy.props.FloatProperty(
        name = "Arrow Width",
        description = "testing",
        default = 2.0,
        min = 0.0,
        soft_max = 12.0,
        subtype = 'FACTOR'
    )
    
    arrow_width2: bpy.props.FloatProperty(
        name = "Arrow Width",
        description = "testing",
        default = 2.0,
        min = 0.0,
        soft_max = 12.0,
        subtype = 'FACTOR'
    )

    dot_scale: bpy.props.FloatProperty(
        name = "Dot Scale",
        default = 0.135,
        min = 0.0,
        soft_max = 1.0,
        subtype = 'FACTOR'
    )
    
    dot_scale2: bpy.props.FloatProperty(
        name = "Dot Scale",
        default = 0.135,
        min = 0.0,
        soft_max = 1.0,
        subtype = 'FACTOR'
    )

    dot_width: bpy.props.FloatProperty(
        name = "Dot Width",
        default = 3.5,
        min = 0.0,
        soft_max = 10.0,
        subtype = 'FACTOR'
    )
    
    dot_width2: bpy.props.FloatProperty(
        name = "Dot Width",
        default = 3.5,
        min = 0.0,
        soft_max = 10.0,
        subtype = 'FACTOR'
    )

    axis_x_color: bpy.props.FloatVectorProperty(
        name = "X Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (1.0, 0.211766, 0.325486),
        min = 0.0,
        max = 1.0
    )
    
    axis_x_color2: bpy.props.FloatVectorProperty(
        name = "X Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (1.0, 0.211766, 0.325486),
        min = 0.0,
        max = 1.0
    )
    
    axis_y_color: bpy.props.FloatVectorProperty(
        name = "Y Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (0.541176, 0.85883, 0.0),
        min = 0.0,
        max = 1.0
    )
    
    axis_y_color2: bpy.props.FloatVectorProperty(
        name = "Y Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (0.541176, 0.85883, 0.0),
        min = 0.0,
        max = 1.0
    )

    axis_z_color: bpy.props.FloatVectorProperty(
        name = "Z Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (0.172551, 0.560786, 1.0),
        min = 0.0,
        max = 1.0
    )
    
    axis_z_color2: bpy.props.FloatVectorProperty(
        name = "Z Axis Color",
        subtype = 'COLOR_GAMMA',
        default = (0.172551, 0.560786, 1.0),
        min = 0.0,
        max = 1.0
    )

    dot_color: bpy.props.FloatVectorProperty(
        name = "Dot Color",
        subtype = 'COLOR_GAMMA',
        default = (0.0, 0.0, 0.0),
        min = 0.0,
        max = 1.0
    )
    
    dot_color2: bpy.props.FloatVectorProperty(
        name = "Dot Color",
        subtype = 'COLOR_GAMMA',
        default = (0.0, 0.0, 0.0),
        min = 0.0,
        max = 1.0
    )
    

    alpha: bpy.props.FloatProperty(
        name = "Alpha",
        default = 1.0,
        min = 0.0,
        max = 1.0,
        subtype = 'FACTOR'
    )
    
    alpha2: bpy.props.FloatProperty(
        name = "Alpha",
        default = 1.0,
        min = 0.0,
        max = 1.0,
        subtype = 'FACTOR'
    )

    alpha_highlight: bpy.props.FloatProperty(
        name = "Alpha Highlight",
        default = 0.75,
        min = 0.0,
        max = 1.0,
        subtype = 'FACTOR'
    )
    
    alpha_highlight2: bpy.props.FloatProperty(
        name = "Alpha Highlight",
        default = 0.75,
        min = 0.0,
        max = 1.0,
        subtype = 'FACTOR'
    )
    

    color_highlight: bpy.props.FloatVectorProperty(
        name = "Highlight Color",
        subtype = 'COLOR_GAMMA',
        default = (1.0, 1.0, 1.0),
        min = 0.0,
        max = 1.0
    )
    
    color_highlight2: bpy.props.FloatVectorProperty(
        name = "Highlight Color",
        subtype = 'COLOR_GAMMA',
        default = (1.0, 1.0, 1.0),
        min = 0.0,
        max = 1.0
    )

# Plus Snap   Plus Snap   Plus Snap   Plus Snap   Plus Snap   Plus Snap   Plus Snap   Plus Snap

    plus_snap_list: bpy.props.StringProperty(
        name = "CURSORPLUS",
        default = 'EDGE'
    )

    plus_snap_align: bpy.props.BoolProperty(
        name = "Align Cursor Rotation",
        default = True
    )      

# Utility props for UI drawing    Utility props for UI drawing    Utility props for UI drawing  
    
    use_parent_plus: bpy.props.BoolProperty(
        name = "Use Parent Panel",
        description = "Define 3D Cursor Plus tab location",
        default = True,
        update = updatePanel
    )

    category_plus: bpy.props.StringProperty(
        name = "Tab Category",
        description = "3D Cursor Plus tab name",
        default = "3D Cursor Plus",
        update = updatePanel
    )
    
    plus_undo: bpy.props.BoolProperty(
        name = "Undo Toggle",
        description = "Enable/Disable Undo for 3D Cursor",
        default = True,
        update = toggleUndo
    )

    edit_pie: bpy.props.BoolProperty(
        name = "Edit Pie Menu",
        default = False,
    )
    
    set_suggest: bpy.props.BoolProperty(
        name = "",
        description = "Click me!",
        default = False,
    )
    
    move_suggest: bpy.props.BoolProperty(
        name = "",
        description = "Click me!",
        default = False,
    )
    
    set_suggest2: bpy.props.BoolProperty(
        name = "",
        description = "Click me!",
        default = False,
    )
    
    move_suggest2: bpy.props.BoolProperty(
        name = "",
        description = "Click me!",
        default = False,
    )
    
    how_it_works: bpy.props.BoolProperty(
        name = "",
        description = "Click me!",
        default = False,
    )    

    preset_select: bpy.props.BoolProperty(
        name = "",
        description = "",
        default = True,
        update = redraw_gizmo
    )
    
    
    def draw(self, context):
        prefs = context.preferences.addons[__package__].preferences
        layout = self.layout
        layout.row().prop(self, 'pref_tabs', expand=True)
        box = layout.box() 

# Gizmo Tab     Gizmo Tab     Gizmo Tab     Gizmo Tab     Gizmo Tab                 

        if self.pref_tabs == 'GIZMO':
            col = box.column()
            #col.use_property_split = True
            col.prop(self, 'use_3d_cursor_gizmo')
            if prefs.use_3d_cursor_gizmo:
                col.separator(type='LINE')
                spl = box.split(factor=0.33)
                # Labels
                colG = spl.column()
                colL = colG.column()
                colL.label(text=" ")
                colL.label(text=" ")
                colL.label(text=" ")

                colL.separator()
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="X:")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Y:")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Z:")
                colG.separator()
                                    
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Axes Size")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Axes Length" )
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Axes Width")
                colL.separator()
                
                colL.separator(type='LINE')
                colL.label(text=" ")
               
                colL.label(text=" ")
                colL.separator()
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Dot Color")
                colL.separator()
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Dot Size")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Dot Line Width")
                 
                colL.separator(type='LINE')
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Gizmo Transparency")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Highlight Color")
                colL = colG.column()
                row = colL.row()
                row.alignment = 'RIGHT'
                row.label(text="Highlight Transparency")
                colL.separator()
                
                # Preset A
                colG = spl.column()
                row = colG.row()
                row.alignment = 'CENTER'
                if self.preset_select:
                    row.label(text="Preset A", icon='CHECKMARK')
                else:
                    row.label(text="Preset A", icon='BLANK1')
                colG.prop(self, 'show_axes')
                
                colG.prop(self, 'axes_gizmo', text="Use Axes as Gizmo")
                
                colG.separator()
                colG.prop(self, 'axis_x_color', text="")
                colG.prop(self, 'axis_y_color', text="")
                colG.prop(self, 'axis_z_color', text="")
                colG.separator()
                                    
                colG.prop(self, 'custom_gizmo_size', text="")
                colG.prop(self, 'arrow_length', text="" )
                colG.prop(self, 'arrow_width', text="")
                colG.separator()
                
                colG.separator(type='LINE')
                colG.prop(self, 'show_dot')
                
                colG.prop(self, 'dot_gizmo', text="Use Dot as Gizmo")
                colG.separator()
                colG.prop(self, 'dot_color', text="")
                colG.separator()
                colG.prop(self, 'dot_scale', text="")
                colG.prop(self, 'dot_width', text="")
 
               
                colG.separator(type='LINE')
                colG.prop(self, 'alpha', text="")
                colG.prop(self, 'color_highlight', text="")
                colG.prop(self, 'alpha_highlight', text="")
                colG.separator()
               
                # Preset B    
                colG = spl.column()
                row = colG.row()
                row.alignment = 'CENTER'
                if self.preset_select:
                    row.label(text="Preset B", icon='BLANK1')
                else:
                    row.label(text="Preset B", icon='CHECKMARK')
                colG.prop(self, 'show_axes2')
                
                colG.prop(self, 'axes_gizmo2', text="Use Axes as Gizmo")
                colG.separator()
                colG.prop(self, 'axis_x_color2', text="")
                colG.prop(self, 'axis_y_color2', text="")
                colG.prop(self, 'axis_z_color2', text="")
                colG.separator()
                                    
                colG.prop(self, 'custom_gizmo_size2', text="")
                colG.prop(self, 'arrow_length2', text="" )
                colG.prop(self, 'arrow_width2', text="")
                colG.separator()
                
                colG.separator(type='LINE')
                colG.prop(self, 'show_dot2')
                
                colG.prop(self, 'dot_gizmo2', text="Use Dot as Gizmo")
                colG.separator()
                colG.prop(self, 'dot_color2', text="")
                colG.separator()
                colG.prop(self, 'dot_scale2', text="")
                colG.prop(self, 'dot_width2', text="")
 
               
                colG.separator(type='LINE')
                colG.prop(self, 'alpha2', text="")
                colG.prop(self, 'color_highlight2', text="")
                colG.prop(self, 'alpha_highlight2', text="")
                colG.separator()

            # Add buttons for saving and loading
            row = layout.row()
            row.operator("cursorplus.save_presets", text="Save Config", icon='FILE_TICK')
            row.operator("cursorplus.load_presets", text="Load Config", icon='FILE_PARENT')     


# Keymap Tab   Keymap Tab   Keymap Tab   Keymap Tab   Keymap Tab   Keymap Tab   Keymap Tab                     

        # 3D Cursor Plus Panel location
        elif self.pref_tabs == 'PLUS_PREFS':
            col = box.column()
            _label_multiline(context=context, text=txt.ops1, parent=col)
            col.separator(type='LINE')
            col.prop(self, 'use_parent_plus', text="Use Default panel location: Sidebar > View > 3D Cursor ")
            if not context.preferences.addons[__package__].preferences.use_parent_plus:
                col.prop(self, 'category_plus', text="Tab category")
            
            box = layout.box()
            
            # Pie and Preset hotkeys            
            wm = bpy.context.window_manager
            kc = wm.keyconfigs.user
            km1 = kc.keymaps.get("3D View")
            km2 = kc.keymaps.get("3D View Tool: Cursor")
            km3 = kc.keymaps.get("Object Mode")
            # Not enabling the pointer until the kmi was modified by user gives both ability
            # to restore it to previous state and prohibit kmi deletion from this tab.
            
            colK = box.column()
            row1a = colK.row()
            for kmi in km1.keymap_items:
                if kmi.name == "3D Cursor Gizmo":
                    if kmi.is_user_modified:
                        row1a.context_pointer_set("keymap", km1) 
                    rna_keymap_ui.draw_kmi([], kc, km1, kmi, row1a, 0)
            row1a.label(text="", icon='BLANK1')
                        
            colK = box.column()
            row1b = colK.row()
            for kmi in km1.keymap_items:
                if kmi.name == "3D Cursor Plus Pie":
                    if kmi.is_user_modified:
                        row1b.context_pointer_set("keymap", km1) 
                    rna_keymap_ui.draw_kmi([], kc, km1, kmi, row1b, 0)
            row1b.prop(self, 'edit_pie', text="", icon='MODIFIER_DATA', emboss=False)
            
	    # PLUS-Ops shortcuts added by user
            for kmi in km3.keymap_items:
                if kmi.idname in plus_ops:
                    colOps = box.column()
                    rowOps = colOps.row()
                    rowOps.context_pointer_set("keymap", km3)
                    rna_keymap_ui.draw_kmi([], kc, km3, kmi, rowOps, 0)
                    rowOps.label(text="", icon='BLANK1')
                    
    
            # Pie menu customization 
            spacer1_X = 0.55
            spacer2_X = 0.7
            all_X = 1.0
            all_Y = 1.2
            if self.edit_pie:

                # Top         
                colK.separator()
                row = colK.row()
                row.alignment = 'CENTER'
                row.scale_x = all_X
                row.scale_y = all_Y
                row.prop(self, "pie_top", text="")

                # Top-Left + Top-Right              
                colK.separator()
                row = colK.row()
                row.alignment = 'CENTER'
                
                row.scale_x = all_X
                row.scale_y = all_Y
                row.prop(self, "pie_top_left", text="")
                spacer = row.row()
                spacer.scale_x = spacer1_X
                spacer.label(text=" ")
                row.prop(self, "pie_top_right", text="")

                # Left + Right            
                colK.separator()
                colK.separator()
                row = colK.row()
                row.alignment = 'CENTER'
                row.scale_x = all_X
                row.scale_y = all_Y
                row.prop(self, "pie_left", text="")
                spacer = row.row()
                spacer.scale_x = spacer2_X
                spacer.label(text=" ")
                row.prop(self, "pie_right", text="")

                # Bottom-Left + Bottom-Right            
                colK.separator()
                colK.separator()
                row = colK.row()
                row.alignment = 'CENTER'
                
                row.scale_x = all_X
                row.scale_y = all_Y
                row.prop(self, "pie_bottom_left", text="")
                spacer = row.row()
                spacer.scale_x = spacer1_X
                spacer.label(text=" ")
                row.prop(self, "pie_bottom_right", text="")

                # Bottom  
                colK.separator()
                row = colK.row()
                row.alignment = 'CENTER'
                row.scale_x = all_X
                row.scale_y = all_Y
                row.prop(self, "pie_bottom", text="")
                colK.separator()
                
                row = colK.row()
                row.operator("cursorplus.save_presets", text="Save Config", icon='FILE_TICK')
                row.operator("cursorplus.load_presets", text="Load Config", icon='FILE_PARENT')

            # Default 3D Cursor related keymap items:    
            layout.label(text="   3D View")
            box = layout.box()
            
            # 'Set 3D Cursor'            
            col2 = box.column()       
            row2 = col2.row()
            for kmi in km1.keymap_items:            
                if kmi.idname == "view3d.cursor3d":
                    if kmi.is_user_modified:
                        row2.context_pointer_set("keymap", km1)
                    rna_keymap_ui.draw_kmi([], kc, km1, kmi, row2, 0)
            row2.prop(self, 'set_suggest', text="", icon='LIGHT', emboss=False)
            if self.set_suggest:
                _label_multiline(context=context, text=txt.header0, parent=col2)
                _label_multiline(context=context, text=txt.set_sug1a, parent=col2)
                _label_multiline(context=context, text=txt.set_sug2, parent=col2)
                col2.separator(type='LINE')  
            
            # 'Move 3D Cursor'             
            col3 = box.column()     
            row3 = col3.row()    
            col3sub = row3.column()
            for kmi in km1.keymap_items:                                    
                if kmi.idname == "transform.translate" and kmi.properties.cursor_transform:
                    if kmi.is_user_modified:
                        col3sub.context_pointer_set("keymap", km1)
                    rna_keymap_ui.draw_kmi([], kc, km1, kmi, col3sub, 0)
            row3.prop(self, 'move_suggest', text="", icon='LIGHT', emboss=False)
            if self.move_suggest:
                _label_multiline(context=context, text=txt.header0, parent=col3)
                _label_multiline(context=context, text=txt.move_sug1, parent=col3)
                _label_multiline(context=context, text=txt.move_sug2, parent=col3)
                _label_multiline(context=context, text=txt.move_sug3, parent=col3)
           

            layout.label(text="   3D View Tool: Cursor")
            box = layout.box()
            
            # 'Tool: Set 3D Cursor'               
            col4 = box.column()
            row4 = col4.row()       
            for kmi in km2.keymap_items:
                if kmi.idname == "view3d.cursor3d":
                    if kmi.is_user_modified:
                            row4.context_pointer_set("keymap", km2)
                    rna_keymap_ui.draw_kmi([], kc, km2, kmi, row4, 0)
            row4.prop(self, 'set_suggest2', text="", icon='LIGHT', emboss=False)
            if self.set_suggest2:
                _label_multiline(context=context, text=txt.header0, parent=col4)
                _label_multiline(context=context, text=txt.set_sug1b, parent=col4)
                _label_multiline(context=context, text=txt.set_sug2, parent=col4)
                col4.separator(type='LINE')

            
            # 'Tool: Move 3D Cursor'    
            col5 = box.column()
            row5 = col5.row()    
            col5sub = row5.column()       
            for kmi in km2.keymap_items:
                if kmi.idname == "transform.translate" and kmi.properties.cursor_transform:
                    if kmi.is_user_modified:
                            col5sub.context_pointer_set("keymap", km2)
                    rna_keymap_ui.draw_kmi([], kc, km2, kmi, col5sub, 0)
            row5.prop(self, 'move_suggest2', text="", icon='LIGHT', emboss=False)
            if self.move_suggest2:
                _label_multiline(context=context, text=txt.header0, parent=col5)
                _label_multiline(context=context, text=txt.move_sug1, parent=col5)
                _label_multiline(context=context, text=txt.move_sug2, parent=col5)
                _label_multiline(context=context, text=txt.move_sug3, parent=col5)
            
            
                    
# Undo Tab   Undo Tab   Undo Tab   Undo Tab   Undo Tab   Undo Tab   Undo Tab   Undo Tab              
            
        else:                       
            colU = box.column()
            if self.how_it_works:
                spl = colU.split(factor=0.7)
            else:
                spl = colU.split(factor=0.95)
            spl.prop(self, 'plus_undo', text="Use the 3D Cursor Undo")
                    
            if self.how_it_works:
                spl.prop(self, 'how_it_works', text=" ", icon='GHOST_ENABLED', emboss=False)
                colU.separator()
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text=" ~ << Cursor Taming Ritual >> ~ ")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="1. Summon a ghost.")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="2. Confine it inside the file.")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="3. Cast a spell that mutually ties it with the cursor.")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="From now on, if one of them moves - the other will follow.")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="Thus, you've obtained a haunted Blender file")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="and a ghost, obeying to Undo and Redo operations.")
                row = colU.row()
                row.alignment = 'CENTER'
                row.label(text="~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                colU.separator()
            else:
                spl.prop(self, 'how_it_works', text="", icon='GHOST_DISABLED', emboss=False)
            colU.separator()

            _label_multiline(context=context, text=txt.undo1, parent=colU)
            colU.label(text="     - 3D Cursor's Location")
            colU.label(text="     - 3D Cursor's Euler Rotation")
            colU.label(text="     - 3D Cursor's Rotation Mode")
            _label_multiline(context=context, text=txt.undo2, parent=colU)
            _label_multiline(context=context, text=txt.undo3, parent=colU)
            _label_multiline(context=context, text=txt.undo4, parent=colU)
            colU.separator()            

 
addon_keymap = []

def register():
    bpy.utils.register_class(CursorPlusPreferences)
    bpy.utils.register_class(CURSORPLUS_OT_gizmo_presets)
     
    kc = bpy.context.window_manager.keyconfigs.addon
    km = kc.keymaps.new(name="3D View", space_type="VIEW_3D")
    kmi = km.keymap_items.new('cursorplus.gizmo_presets', 'D', 'PRESS', shift=True, ctrl=True)
    addon_keymap.append((km,kmi))
    kmi = km.keymap_items.new('wm.call_menu_pie', 'D', 'PRESS', ctrl=True)
    kmi.properties.name = "CURSORPLUS_MT_cursor_plus_pie"
    addon_keymap.append((km,kmi))

    
def unregister():
    bpy.utils.unregister_class(CURSORPLUS_OT_gizmo_presets)
    bpy.utils.unregister_class(CursorPlusPreferences)
        
    for km,kmi in addon_keymap:
        km.keymap_items.remove(kmi)
    addon_keymap.clear()